"""Chaos — Retry logic: intermittent failures, backoff timing, connection errors."""

from __future__ import annotations

from unittest.mock import patch

import httpx
import pytest
import respx

from kanoniv.client.client import KanonivAsyncClient, KanonivClient
from kanoniv.exceptions import (
    KanonivError,
    NotFoundError,
    RateLimitError,
    ServerError,
    ValidationError,
)

from .conftest import API_KEY, BASE_URL


# ---------------------------------------------------------------------------
# Retryable status codes
# ---------------------------------------------------------------------------

class TestRetryableStatusCodes:
    """Retryable codes (408/429/502/503/504) are retried; others are not."""

    @pytest.mark.parametrize("status_code", [408, 429, 502, 503, 504])
    @patch("kanoniv.client._transport.time.sleep")
    def test_retryable_then_success(
        self, mock_sleep, mock_api: respx.Router, status_code: int,
    ):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=[
                    httpx.Response(status_code, json={"error": "transient"}),
                    httpx.Response(200, json={"total": 42}),
                ],
            )
            result = retry_client.stats()
            assert result["total"] == 42
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_400_not_retried(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            route = mock_api.get("/v1/stats").mock(
                return_value=httpx.Response(400, json={"error": "bad"}),
            )
            with pytest.raises(ValidationError):
                retry_client.stats()
            assert len(route.calls) == 1
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_404_not_retried(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            route = mock_api.get("/v1/stats").mock(
                return_value=httpx.Response(404, json={"error": "nope"}),
            )
            with pytest.raises(NotFoundError):
                retry_client.stats()
            assert len(route.calls) == 1
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_all_retries_exhausted(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                return_value=httpx.Response(503, json={"error": "down"}),
            )
            with pytest.raises(ServerError):
                retry_client.stats()
        finally:
            retry_client.close()

    @pytest.mark.asyncio
    @patch("asyncio.sleep")
    async def test_async_429_retried(self, mock_sleep, mock_api: respx.Router):
        ac = KanonivAsyncClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=[
                    httpx.Response(429, json={"error": "rate limited"}),
                    httpx.Response(200, json={"total": 99}),
                ],
            )
            result = await ac.stats()
            assert result["total"] == 99
        finally:
            await ac.close()


# ---------------------------------------------------------------------------
# Backoff timing
# ---------------------------------------------------------------------------

class TestBackoffTiming:
    """Exponential backoff and Retry-After header."""

    @patch("kanoniv.client._transport.time.sleep")
    def test_exponential_backoff(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=[
                    httpx.Response(503, json={"error": "down"}),
                    httpx.Response(503, json={"error": "down"}),
                    httpx.Response(200, json={"total": 1}),
                ],
            )
            retry_client.stats()
            calls = mock_sleep.call_args_list
            assert calls[0][0][0] == pytest.approx(0.5)  # 0.5 * 2^0
            assert calls[1][0][0] == pytest.approx(1.0)  # 0.5 * 2^1
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_retry_after_header_respected(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=[
                    httpx.Response(
                        429,
                        json={"error": "slow down"},
                        headers={"Retry-After": "3"},
                    ),
                    httpx.Response(200, json={"total": 1}),
                ],
            )
            retry_client.stats()
            assert mock_sleep.call_args_list[0][0][0] == pytest.approx(3.0)
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_invalid_retry_after_falls_back(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=[
                    httpx.Response(
                        429,
                        json={"error": "slow down"},
                        headers={"Retry-After": "not-a-number"},
                    ),
                    httpx.Response(200, json={"total": 1}),
                ],
            )
            retry_client.stats()
            assert mock_sleep.call_args_list[0][0][0] == pytest.approx(0.5)
        finally:
            retry_client.close()


# ---------------------------------------------------------------------------
# Connection failures
# ---------------------------------------------------------------------------

class TestConnectionFailures:
    """ConnectError and ReadTimeout are retried then raise KanonivError."""

    @patch("kanoniv.client._transport.time.sleep")
    def test_connect_error_retried_then_raises(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=httpx.ConnectError("connection refused"),
            )
            with pytest.raises(KanonivError, match="Connection failed"):
                retry_client.stats()
            assert mock_sleep.call_count == 2
        finally:
            retry_client.close()

    @patch("kanoniv.client._transport.time.sleep")
    def test_read_timeout_retried_then_raises(self, mock_sleep, mock_api: respx.Router):
        retry_client = KanonivClient(api_key=API_KEY, base_url=BASE_URL, max_retries=2)
        try:
            mock_api.get("/v1/stats").mock(
                side_effect=httpx.ReadTimeout("read timed out"),
            )
            with pytest.raises(KanonivError, match="Connection failed"):
                retry_client.stats()
            assert mock_sleep.call_count == 2
        finally:
            retry_client.close()



"""Tests for PandasAdapter (guarded by pandas availability)."""
import pytest

pd = pytest.importorskip("pandas")

from kanoniv.adapters.pandas import PandasAdapter
from kanoniv.source import Source


class TestPandasAdapter:
    def test_schema_type_inference(self):
        df = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["Alice", "Bob", "Charlie"],
                "score": [9.5, 8.0, 7.2],
                "active": [True, False, True],
            }
        )
        adapter = PandasAdapter(df)
        schema = adapter.schema()
        col_map = {c.name: c for c in schema.columns}
        assert col_map["id"].dtype == "number"
        assert col_map["name"].dtype == "string"
        assert col_map["score"].dtype == "number"
        assert col_map["active"].dtype == "boolean"

    def test_schema_datetime_inference(self):
        df = pd.DataFrame(
            {
                "created_at": pd.to_datetime(["2024-01-01", "2024-02-01"]),
            }
        )
        adapter = PandasAdapter(df)
        schema = adapter.schema()
        assert schema.columns[0].dtype == "date"

    def test_nan_handling(self):
        df = pd.DataFrame(
            {
                "name": ["Alice", None, "Charlie"],
                "score": [1.0, float("nan"), 3.0],
            }
        )
        adapter = PandasAdapter(df)
        rows = list(adapter.iter_rows())
        assert rows[1]["name"] == ""
        assert rows[1]["score"] == ""

    def test_iter_rows_matches_dataframe(self):
        df = pd.DataFrame(
            {
                "id": [10, 20],
                "name": ["X", "Y"],
            }
        )
        adapter = PandasAdapter(df)
        rows = list(adapter.iter_rows())
        assert len(rows) == 2
        assert rows[0]["id"] == "10"
        assert rows[1]["name"] == "Y"

    def test_row_count(self):
        df = pd.DataFrame({"a": range(50)})
        adapter = PandasAdapter(df)
        assert adapter.row_count() == 50

    def test_source_from_pandas(self):
        df = pd.DataFrame(
            {"id": [1, 2], "email": ["a@b.com", "c@d.com"]}
        )
        src = Source.from_pandas("test", df, primary_key="id")
        entities = src.to_entities("customer")
        assert len(entities) == 2
        assert entities[0]["data"]["email"] == "a@b.com"

    def test_invalid_input_raises(self):
        with pytest.raises(TypeError, match="DataFrame"):
            PandasAdapter({"not": "a dataframe"})

    def test_nullable_detection(self):
        df = pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": [1, None, 3],
            }
        )
        adapter = PandasAdapter(df)
        schema = adapter.schema()
        col_map = {c.name: c for c in schema.columns}
        assert col_map["a"].nullable is False
        assert col_map["b"].nullable is True



"""Rules resource — list active, create version, history."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .._transport import AsyncTransport, SyncTransport


class RulesResource:
    def __init__(self, transport: SyncTransport) -> None:
        self._t = transport

    def list(self) -> list[dict[str, Any]]:
        return self._t.request("GET", "/v1/rules")

    def create(
        self,
        *,
        name: str,
        rule_type: str,
        config: dict[str, Any],
        weight: float | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        body: dict[str, Any] = {"name": name, "rule_type": rule_type, "config": config, **kwargs}
        if weight is not None:
            body["weight"] = weight
        return self._t.request("POST", "/v1/rules", json=body)

    def history(self, name: str) -> list[dict[str, Any]]:
        return self._t.request("GET", f"/v1/rules/{name}/history")


class AsyncRulesResource:
    def __init__(self, transport: AsyncTransport) -> None:
        self._t = transport

    async def list(self) -> list[dict[str, Any]]:
        return await self._t.request("GET", "/v1/rules")

    async def create(
        self,
        *,
        name: str,
        rule_type: str,
        config: dict[str, Any],
        weight: float | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        body: dict[str, Any] = {"name": name, "rule_type": rule_type, "config": config, **kwargs}
        if weight is not None:
            body["weight"] = weight
        return await self._t.request("POST", "/v1/rules", json=body)

    async def history(self, name: str) -> list[dict[str, Any]]:
        return await self._t.request("GET", f"/v1/rules/{name}/history")
